package ru.dbotthepony.mc.otm.block.entity.decorative

import net.minecraft.advancements.CriteriaTriggers
import net.minecraft.core.BlockPos
import net.minecraft.core.HolderLookup
import net.minecraft.core.registries.Registries
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.Tag
import net.minecraft.network.chat.Component
import net.minecraft.resources.ResourceKey
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.level.ServerLevel
import net.minecraft.server.level.ServerPlayer
import net.minecraft.sounds.SoundSource
import net.minecraft.world.Containers
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.gameevent.GameEvent
import net.minecraft.world.level.storage.loot.LootParams
import net.minecraft.world.level.storage.loot.LootTable
import net.minecraft.world.level.storage.loot.parameters.LootContextParamSets
import net.minecraft.world.level.storage.loot.parameters.LootContextParams
import net.minecraft.world.phys.Vec3
import net.neoforged.neoforge.capabilities.Capabilities
import ru.dbotthepony.mc.otm.block.decorative.CargoCrateBlock
import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.HandlerFilter
import ru.dbotthepony.mc.otm.container.slotted.FilteredContainerSlot
import ru.dbotthepony.mc.otm.container.slotted.SlottedContainer
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.random
import ru.dbotthepony.mc.otm.menu.decorative.CargoCrateMenu
import ru.dbotthepony.mc.otm.registry.game.MBlockEntities
import ru.dbotthepony.mc.otm.registry.game.MSoundEvents

class CargoCrateBlockEntity(
	p_155229_: BlockPos,
	p_155230_: BlockState
) : MatteryDeviceBlockEntity(MBlockEntities.CARGO_CRATE, p_155229_, p_155230_) {
	private inner class Slot(container: SlottedContainer, slot: Int) : FilteredContainerSlot(container, slot) {
		override fun canAutomationPlaceItem(itemStack: ItemStack): Boolean {
			return super<FilteredContainerSlot>.canAutomationPlaceItem(itemStack) && lootTable == null
		}

		override fun canAutomationTakeItem(desired: Int): Boolean {
			return super.canAutomationTakeItem(desired) && lootTable == null
		}
	}

	val container = SlottedContainer.Builder()
		.add(CAPACITY, ::Slot)
		.onChanged(::setChanged)
		.build()
		.also(::addDroppableContainer)

	private var interactingPlayers = 0

	override fun beforeDroppingItems(oldBlockState: BlockState, level: Level, blockPos: BlockPos, newBlockState: BlockState, movedByPiston: Boolean) {
		unpackLootTable()
	}

	override fun beforeDestroyedByPlayer(level: Level, blockPos: BlockPos, blockState: BlockState, player: Player) {
		unpackLootTable(player)
	}

	var lootTable: ResourceKey<LootTable>? = null
	var lootTableSeed: Long? = null

	fun onPlayerOpen() {
		val level = level

		if (interactingPlayers++ == 0 && level != null && !isRemoved && level.getBlockState(blockPos).block is CargoCrateBlock) {
			level.setBlock(blockPos, blockState.setValue(CargoCrateBlock.IS_OPEN, true), Block.UPDATE_CLIENTS)
			level.playSound(null, blockPos, MSoundEvents.CARGO_CRATE_OPEN, SoundSource.BLOCKS, 1f, 0.8f + level.random.nextFloat() * 0.2f)
			level.gameEvent(GameEvent.CONTAINER_OPEN, blockPos, GameEvent.Context.of(blockState))
		}
	}

	fun onPlayerClose() {
		val level = level

		if (--interactingPlayers == 0 && level != null && !isRemoved && level.getBlockState(blockPos).block is CargoCrateBlock) {
			level.setBlock(blockPos, blockState.setValue(CargoCrateBlock.IS_OPEN, false), Block.UPDATE_CLIENTS)
			level.gameEvent(GameEvent.CONTAINER_CLOSE, blockPos, GameEvent.Context.of(blockState))
		}
	}

	init {
		exposeGlobally(Capabilities.ItemHandler.BLOCK, container)
		savetablesLevel.stateful(::container, INVENTORY_KEY)
	}

	override fun saveLevel(nbt: CompoundTag, registry: HolderLookup.Provider) {
		super.saveLevel(nbt, registry)

		if (lootTable != null) {
			nbt.putString(LOOT_TABLE_KEY, lootTable!!.location().toString())
			nbt.putLong(LOOT_TABLE_SEED_KEY, lootTableSeed ?: 0L)
		}
	}

	override fun loadAdditional(nbt: CompoundTag, registry: HolderLookup.Provider) {
		super.loadAdditional(nbt, registry)

		if (nbt.contains(LOOT_TABLE_KEY, Tag.TAG_STRING.toInt())) {
			lootTable = ResourceLocation.tryParse(nbt.getString(LOOT_TABLE_KEY))?.let { ResourceKey.create(Registries.LOOT_TABLE, it) }
			lootTableSeed = if (nbt.contains(LOOT_TABLE_SEED_KEY, Tag.TAG_LONG.toInt())) nbt.getLong(LOOT_TABLE_SEED_KEY) else 0L
		}
	}

	fun unpackLootTable(ply: Player? = null) {
		val lootTable = lootTable ?: return
		val lootTableSeed = lootTableSeed ?: 0L
		val server = level?.server ?: return

		val loot = server.reloadableRegistries().getLootTable(lootTable)

		if (ply is ServerPlayer) {
			CriteriaTriggers.GENERATE_LOOT.trigger(ply, lootTable)
		}

		this.lootTable = null
		this.lootTableSeed = null

		val params = LootParams.Builder(level as ServerLevel)
			.withParameter(LootContextParams.ORIGIN, Vec3.atCenterOf(this.worldPosition))

		if (ply != null) {
			params.withLuck(ply.luck).withParameter(LootContextParams.THIS_ENTITY, ply)
		}

		Containers.dropContents(level as ServerLevel, blockPos, container)
		loot.fill(container, params.create(LootContextParamSets.CHEST), lootTableSeed)
	}

	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu? {
		if (lootTable != null && ply.isSpectator)
			return null

		unpackLootTable(ply)
		return CargoCrateMenu(containerID, inventory, this)
	}

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.cargo_crate")
		const val CAPACITY = 9 * 6
	}
}
